package com.sun.org.apache.xerces.internal.impl.xs;

import java.util.ArrayList;
import java.util.List;
import com.sun.org.apache.xerces.internal.xs.XSObjectList;
import com.sun.org.apache.xerces.internal.xs.XSSimpleTypeDefinition;
import com.sun.org.apache.xerces.internal.xs.XSTypeDefinition;
import com.sun.org.apache.xerces.internal.xni.QName;
import java.util.HashMap;
import java.util.Map;

public class SubstitutionGroupHandler
{
    private static final XSElementDecl[] EMPTY_GROUP;
    private final XSElementDeclHelper fXSElementDeclHelper;
    Map<XSElementDecl, Object> fSubGroupsB;
    private static final SubstitutionGroupHandler.OneSubGroup[] EMPTY_VECTOR;
    Map<XSElementDecl, XSElementDecl[]> fSubGroups;
    
    public SubstitutionGroupHandler(XSElementDeclHelper elementDeclHelper) {
        this.fSubGroupsB = new HashMap();
        this.fSubGroups = new HashMap();
        this.fXSElementDeclHelper = elementDeclHelper;
    }
    
    public XSElementDecl getMatchingElemDecl(QName element, XSElementDecl exemplar) {
        if (element.localpart == exemplar.fName && element.uri == exemplar.fTargetNamespace) {
            return exemplar;
        }
        if (exemplar.fScope != 1) {
            return null;
        }
        if ((exemplar.fBlock & 0x4) != 0x0) {
            return null;
        }
        XSElementDecl eDecl = this.fXSElementDeclHelper.getGlobalElementDecl(element);
        if (eDecl == null) {
            return null;
        }
        if (this.substitutionGroupOK(eDecl, exemplar, exemplar.fBlock)) {
            return eDecl;
        }
        return null;
    }
    
    protected boolean substitutionGroupOK(XSElementDecl element, XSElementDecl exemplar, short blockingConstraint) {
        if (element == exemplar) {
            return true;
        }
        if ((blockingConstraint & 0x4) != 0x0) {
            return false;
        }
        XSElementDecl subGroup;
        for (subGroup = element.fSubGroup; subGroup != null && subGroup != exemplar; subGroup = subGroup.fSubGroup) {}
        return subGroup != null && this.typeDerivationOK(element.fType, exemplar.fType, blockingConstraint);
    }
    
    private boolean typeDerivationOK(XSTypeDefinition derived, XSTypeDefinition base, short blockingConstraint) {
        short devMethod = 0;
        short blockConstraint = blockingConstraint;
        XSTypeDefinition type = derived;
        while (type != base && type != SchemaGrammar.fAnyType) {
            if (type.getTypeCategory() == 15) {
                devMethod |= ((XSComplexTypeDecl)type).fDerivedBy;
            }
            else {
                devMethod |= 0x2;
            }
            type = type.getBaseType();
            if (type == null) {
                type = SchemaGrammar.fAnyType;
            }
            if (type.getTypeCategory() != 15) {
                continue;
            }
            blockConstraint |= ((XSComplexTypeDecl)type).fBlock;
        }
        if (type != base) {
            if (base.getTypeCategory() == 16) {
                XSSimpleTypeDefinition st = (XSSimpleTypeDefinition)base;
                if (st.getVariety() == 3) {
                    XSObjectList memberTypes = st.getMemberTypes();
                    int length = memberTypes.getLength();
                    int i = 0;
                    while (i < length) {
                        if (this.typeDerivationOK(derived, (XSTypeDefinition)memberTypes.item(i), blockingConstraint)) {
                            return true;
                        }
                        ++i;
                    }
                }
            }
            return false;
        }
        else {
            return (devMethod & blockConstraint) == 0x0;
        }
    }
    
    public boolean inSubstitutionGroup(XSElementDecl element, XSElementDecl exemplar) {
        return this.substitutionGroupOK(element, exemplar, exemplar.fBlock);
    }
    
    public void reset() {
        this.fSubGroupsB.clear();
        this.fSubGroups.clear();
    }
    
    public void addSubstitutionGroup(XSElementDecl[] elements) {
        for (int i = elements.length - 1; i >= 0; --i) {
            XSElementDecl element = elements[i];
            XSElementDecl subHead = element.fSubGroup;
            List<XSElementDecl> subGroup = (List)this.fSubGroupsB.get(subHead);
            if (subGroup == null) {
                subGroup = new ArrayList();
                this.fSubGroupsB.put(subHead, subGroup);
            }
            subGroup.add(element);
        }
    }
    
    public XSElementDecl[] getSubstitutionGroup(XSElementDecl element) {
        XSElementDecl[] subGroup = (XSElementDecl[])this.fSubGroups.get(element);
        if (subGroup != null) {
            return subGroup;
        }
        if ((element.fBlock & 0x4) != 0x0) {
            this.fSubGroups.put(element, SubstitutionGroupHandler.EMPTY_GROUP);
            return SubstitutionGroupHandler.EMPTY_GROUP;
        }
        SubstitutionGroupHandler.OneSubGroup[] groupB = this.getSubGroupB(element, new SubstitutionGroupHandler.OneSubGroup());
        int len = groupB.length;
        int rlen = 0;
        XSElementDecl[] ret = new XSElementDecl[len];
        for (int i = 0; i < len; ++i) {
            if ((element.fBlock & groupB[i].dMethod) == 0x0) {
                ret[rlen++] = groupB[i].sub;
            }
        }
        if (rlen < len) {
            XSElementDecl[] ret2 = new XSElementDecl[rlen];
            System.arraycopy(ret, 0, ret2, 0, rlen);
            ret = ret2;
        }
        this.fSubGroups.put(element, ret);
        return ret;
    }
    
    private SubstitutionGroupHandler.OneSubGroup[] getSubGroupB(XSElementDecl element, SubstitutionGroupHandler.OneSubGroup methods) {
        Object subGroup = this.fSubGroupsB.get(element);
        if (subGroup == null) {
            this.fSubGroupsB.put(element, SubstitutionGroupHandler.EMPTY_VECTOR);
            return SubstitutionGroupHandler.EMPTY_VECTOR;
        }
        if ((subGroup instanceof SubstitutionGroupHandler.OneSubGroup[])) {
            return (SubstitutionGroupHandler.OneSubGroup[])(SubstitutionGroupHandler.OneSubGroup[])subGroup;
        }
        List<XSElementDecl> group = (ArrayList)subGroup;
        List<SubstitutionGroupHandler.OneSubGroup> newGroup = new ArrayList();
        for (int i = group.size() - 1; i >= 0; --i) {
            XSElementDecl sub = (XSElementDecl)group.get(i);
            if (this.getDBMethods(sub.fType, element.fType, methods)) {
                short dMethod = methods.dMethod;
                short bMethod = methods.bMethod;
                newGroup.add(new SubstitutionGroupHandler.OneSubGroup(sub, methods.dMethod, methods.bMethod));
                SubstitutionGroupHandler.OneSubGroup[] group2 = this.getSubGroupB(sub, methods);
                for (int j = group2.length - 1; j >= 0; --j) {
                    short dSubMethod = (short)(dMethod | group2[j].dMethod);
                    short bSubMethod = (short)(bMethod | group2[j].bMethod);
                    if ((dSubMethod & bSubMethod) == 0x0) {
                        newGroup.add(new SubstitutionGroupHandler.OneSubGroup(group2[j].sub, dSubMethod, bSubMethod));
                    }
                }
            }
        }
        SubstitutionGroupHandler.OneSubGroup[] ret = new SubstitutionGroupHandler.OneSubGroup[newGroup.size()];
        for (int k = newGroup.size() - 1; k >= 0; --k) {
            ret[k] = (SubstitutionGroupHandler.OneSubGroup)newGroup.get(k);
        }
        this.fSubGroupsB.put(element, ret);
        return ret;
    }
    
    private boolean getDBMethods(XSTypeDefinition typed, XSTypeDefinition typeb, SubstitutionGroupHandler.OneSubGroup methods) {
        short dMethod = 0;
        short bMethod = 0;
        while (typed != typeb && typed != SchemaGrammar.fAnyType) {
            if (typed.getTypeCategory() == 15) {
                dMethod |= ((XSComplexTypeDecl)typed).fDerivedBy;
            }
            else {
                dMethod |= 0x2;
            }
            typed = typed.getBaseType();
            if (typed == null) {
                typed = SchemaGrammar.fAnyType;
            }
            if (typed.getTypeCategory() != 15) {
                continue;
            }
            bMethod |= ((XSComplexTypeDecl)typed).fBlock;
        }
        if (typed != typeb || (dMethod & bMethod) != 0x0) {
            return false;
        }
        methods.dMethod = dMethod;
        methods.bMethod = bMethod;
        return true;
    }
    
    static {
        EMPTY_GROUP = new XSElementDecl[0];
        EMPTY_VECTOR = new SubstitutionGroupHandler.OneSubGroup[0];
    }
    
    private static final class OneSubGroup
    {
        XSElementDecl sub;
        short dMethod;
        short bMethod;
        
        OneSubGroup() {
        }
        
        OneSubGroup(XSElementDecl sub, short dMethod, short bMethod) {
            this.sub = sub;
            this.dMethod = dMethod;
            this.bMethod = bMethod;
        }
    }
}
